<!DOCTYPE html>
<!-- DO NOT EDIT! This test has been generated by /html/canvas/tools/gentest.py. -->
<title>Canvas test: {{ name }}</title>
<h1 style="font-size: 20px;">{{ name }}</h1>
<p class="desc">{{ desc }}</p>
{% if notes %}<p class="notes">{{ notes }}{% endif %}
{% for image in images %}
<img src="/images/{{ image }}" id="{{ image }}" class="resource">
{% endfor %}
{% for svgimage in svgimages %}
<svg><image xlink:href="/images/{{ svgimage
      }}" id="{{ svgimage }}" class="resource"></svg>
{% endfor %}

<div style="display: grid; grid-gap: 4px;
            grid-template-columns: repeat({{ grid_width }}, max-content);
            font-size: 13px; text-align: center;">
{% for variant in element_variants %}
<span>
  {% for variant_name in variant.grid_variant_names %}
  <div>{{ variant_name }}</div>
  {% endfor %}
  <div style="width: {{ variant.size[0] }}px; height: {{ variant.size[1]
              }}px; outline: 1px solid">
    {{ variant.html_reference | trim | indent(4) }}
  </div>
</span>

{% endfor %}
</div>
